/* OggDec
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Copyright 2002, John Edwards
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#ifdef _WIN32
#include <io.h>
#endif
#include <fcntl.h>
#include <math.h>
#include "audio.h"


audio_file *open_output_audio_file(char *infile, int samplerate, int channels,
                            int outputFormat, int fileType, unsigned int knownlength)
{
	audio_file *aufile = malloc(sizeof(audio_file));

	aufile->outputFormat = outputFormat;
	aufile->samplerate = samplerate;
	aufile->channels = channels;
	aufile->fileType = fileType;
	if(knownlength)
		aufile->samples = knownlength;
	else
		aufile->samples = 0;

	switch (outputFormat)
	{
		case WAV_FMT_8BIT:
			aufile->bits_per_sample = 8;
			break;
		case WAV_FMT_16BIT:
			aufile->bits_per_sample = 16;
			break;
		case WAV_FMT_24BIT:
			aufile->bits_per_sample = 24;
			break;
		case WAV_FMT_32BIT:
		case WAV_FMT_FLOAT:
			aufile->bits_per_sample = 32;
			break;
		default:
			if (aufile)
				free(aufile);
			return NULL;
	}

	if(!infile) {
#ifdef __BORLANDC__
		setmode(fileno(stdout), O_BINARY);
#elif defined _MSC_VER || defined __MINGW32__
		_setmode(_fileno(stdout), _O_BINARY);
#endif
		aufile->sndfile = stdout;
	}
	else
		aufile->sndfile = fopen(infile, "wb");

	if (aufile->sndfile == NULL){
		if (aufile)
			free(aufile);
		return NULL;
	}

	if (aufile->fileType == OUTPUT_WAV) {
		if (aufile->channels > 2)
			write_wav_extensible_header(aufile);
		else
			write_wav_header(aufile);
		aufile->samples = 0;
	}

	return aufile;
}

int write_audio_file(audio_file *aufile, void *sample_buffer, int samples)
{
	switch (aufile->outputFormat) {
		case WAV_FMT_8BIT:
			return write_audio_8bit(aufile, sample_buffer, samples);
		case WAV_FMT_16BIT:
			return write_audio_16bit(aufile, sample_buffer, samples);
		case WAV_FMT_24BIT:
			return write_audio_24bit(aufile, sample_buffer, samples);
		case WAV_FMT_32BIT:
			return write_audio_32bit(aufile, sample_buffer, samples);
		case WAV_FMT_FLOAT:
			return write_audio_float(aufile, sample_buffer, samples);
		default:
			return 0;
	}

	return 0;
}

void close_audio_file(audio_file *aufile, char *fileName)
{
	if (aufile->fileType == OUTPUT_WAV && fileName) {
		fseek(aufile->sndfile, 0, SEEK_SET);
		if (aufile->channels > 2)
			write_wav_extensible_header(aufile);
		else
			write_wav_header(aufile);
	}

	fclose(aufile->sndfile);

	if (aufile)
		free(aufile);
}

static int write_wav_header(audio_file *aufile)
{
	unsigned char  header[44],
	               *p         = header;
	unsigned int   bytes      = (aufile->bits_per_sample + 7) / 8;
	double         data_size;
	unsigned long  word32;

	if(aufile->samples)
		data_size  = aufile->samples * bytes;
	else
		data_size = 0x7fffffff;

	*p++ = 'R';
	*p++ = 'I';
	*p++ = 'F';
	*p++ = 'F';

	word32 = data_size + (44 - 8) < (double)MAXWAVESIZE ?
		(unsigned long)data_size + (44 - 8)  :  (unsigned long)MAXWAVESIZE;
	*p++ = (unsigned char)((word32 >>  0)&0xff);
	*p++ = (unsigned char)((word32 >>  8)&0xff);
	*p++ = (unsigned char)((word32 >> 16)&0xff);
	*p++ = (unsigned char)((word32 >> 24)&0xff);

	*p++ = 'W';
	*p++ = 'A';
	*p++ = 'V';
	*p++ = 'E';

	*p++ = 'f';
	*p++ = 'm';
	*p++ = 't';
	*p++ = ' ';

	*p++ = 0x10;
	*p++ = 0x00;
	*p++ = 0x00;
	*p++ = 0x00;

	if (aufile->outputFormat == WAV_FMT_FLOAT){
		*p++ = 0x03;
		*p++ = 0x00;
	}
	else{
		*p++ = 0x01;
		*p++ = 0x00;
	}

	*p++ = (unsigned char)((aufile->channels >> 0)&0xff);
	*p++ = (unsigned char)((aufile->channels >> 8)&0xff);

	word32 = (unsigned long) (aufile->samplerate + 0.5);
	*p++ = (unsigned char)((word32 >>  0)&0xff);
	*p++ = (unsigned char)((word32 >>  8)&0xff);
	*p++ = (unsigned char)((word32 >> 16)&0xff);
	*p++ = (unsigned char)((word32 >> 24)&0xff);

	word32 *= bytes * aufile->channels;
	*p++ = (unsigned char)((word32 >>  0)&0xff);
	*p++ = (unsigned char)((word32 >>  8)&0xff);
	*p++ = (unsigned char)((word32 >> 16)&0xff);
	*p++ = (unsigned char)((word32 >> 24)&0xff);

	word32 = bytes * aufile->channels;
	*p++ = (unsigned char)((word32 >>  0)&0xff);
	*p++ = (unsigned char)((word32 >>  8)&0xff);

	*p++ = (unsigned char)((aufile->bits_per_sample >> 0)&0xff);
	*p++ = (unsigned char)((aufile->bits_per_sample >> 8)&0xff);

	*p++ = 'd';
	*p++ = 'a';
	*p++ = 't';
	*p++ = 'a';

	word32 = data_size < MAXWAVESIZE ? (unsigned long)data_size : (unsigned long)MAXWAVESIZE;
	*p++ = (unsigned char)((word32 >>  0)&0xff);
	*p++ = (unsigned char)((word32 >>  8)&0xff);
	*p++ = (unsigned char)((word32 >> 16)&0xff);
	*p++ = (unsigned char)((word32 >> 24)&0xff);

	return fwrite(header, sizeof(header), 1, aufile->sndfile);
}

static int write_wav_extensible_header(audio_file *aufile)
{
	unsigned char header[68];
	unsigned char* p = header;
	unsigned int bytes = (aufile->bits_per_sample + 7) / 8;
	float data_size = (float)bytes * aufile->samples;
	unsigned long channelMask;
	unsigned long word32;

	if (aufile->channels == 3)
		channelMask = 7;				// 0x7
	else if (aufile->channels == 4)
		channelMask = 51;				// 0x33
	else if (aufile->channels == 5)
		channelMask = 55;				// 0x37
	else if (aufile->channels == 6)
		channelMask = 63;				// 0x3F
	else if (aufile->channels == 7)
		channelMask = 1807;				// 0x70F
	else if (aufile->channels == 8)
		channelMask = 1599;				// 0x63F
	else
		channelMask = 0;

	*p++ = 'R'; *p++ = 'I'; *p++ = 'F'; *p++ = 'F';

	word32 = (data_size + (68 - 8) < (float)MAXWAVESIZE) ?
        	(unsigned long)data_size + (68 - 8)  :  (unsigned long)MAXWAVESIZE;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	*p++ = 'W'; *p++ = 'A'; *p++ = 'V'; *p++ = 'E';

	*p++ = 'f'; *p++ = 'm'; *p++ = 't'; *p++ = ' ';

	*p++ = /*0x10*/0x28; *p++ = 0x00; *p++ = 0x00; *p++ = 0x00;

	/* WAVE_FORMAT_EXTENSIBLE */
	*p++ = 0xFE; *p++ = 0xFF;

	*p++ = (unsigned char)(aufile->channels >> 0);
	*p++ = (unsigned char)(aufile->channels >> 8);

	word32 = (unsigned long)(aufile->samplerate + 0.5);
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	word32 = aufile->samplerate * bytes * aufile->channels;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	word32 = bytes * aufile->channels;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);

	*p++ = (unsigned char)(aufile->bits_per_sample >> 0);
	*p++ = (unsigned char)(aufile->bits_per_sample >> 8);

	/* cbSize */
	*p++ = (unsigned char)(22);
	*p++ = (unsigned char)(0);

	/* WAVEFORMATEXTENSIBLE */

	/* wValidBitsPerSample */
	*p++ = (unsigned char)(aufile->bits_per_sample >> 0);
	*p++ = (unsigned char)(aufile->bits_per_sample >> 8);

	/* dwChannelMask */
	word32 = channelMask;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	/* SubFormat */
	if (aufile->outputFormat == WAV_FMT_FLOAT)
	{
		/* KSDATAFORMAT_SUBTYPE_IEEE_FLOAT: 00000003-0000-0010-8000-00aa00389b71 */
		*p++ = 0x03;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00; *p++ = 0x00; *p++ = 0x10; *p++ = 0x00; *p++ = 0x80; *p++ = 0x00;
		*p++ = 0x00; *p++ = 0xaa; *p++ = 0x00; *p++ = 0x38; *p++ = 0x9b; *p++ = 0x71;
	} else {
		/* KSDATAFORMAT_SUBTYPE_PCM: 00000001-0000-0010-8000-00aa00389b71 */
		*p++ = 0x01;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00;
		*p++ = 0x00; *p++ = 0x00; *p++ = 0x10; *p++ = 0x00; *p++ = 0x80; *p++ = 0x00;
		*p++ = 0x00; *p++ = 0xaa; *p++ = 0x00; *p++ = 0x38; *p++ = 0x9b; *p++ = 0x71;
	}

	/* end WAVEFORMATEXTENSIBLE */

	*p++ = 'd'; *p++ = 'a'; *p++ = 't'; *p++ = 'a';

	word32 = data_size < MAXWAVESIZE ?
        	(unsigned long)data_size : (unsigned long)MAXWAVESIZE;
	*p++ = (unsigned char)(word32 >>  0);
	*p++ = (unsigned char)(word32 >>  8);
	*p++ = (unsigned char)(word32 >> 16);
	*p++ = (unsigned char)(word32 >> 24);

	return fwrite(header, sizeof(header), 1, aufile->sndfile);
}

static int write_audio_8bit(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int              ret;
	unsigned int     i;
	unsigned char    *sample_buffer8 = (unsigned char*)sample_buffer;
	unsigned char    *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++)
		data[i] = (sample_buffer8[i]+128) & 0xFF;

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_16bit(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int          ret;
	unsigned int i;
	short        *sample_buffer16 = (short*)sample_buffer;
	char         *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++){
		data[i*2] = (char)(sample_buffer16[i] & 0xFF);
		data[i*2+1] = (char)((sample_buffer16[i] >> 8) & 0xFF);
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_24bit(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int          ret;
	unsigned int i;
	long         *sample_buffer24 = (long*)sample_buffer;
	char         *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++){
		data[i*3] = (char)(sample_buffer24[i] & 0xFF);
		data[i*3+1] = (char)((sample_buffer24[i] >> 8) & 0xFF);
		data[i*3+2] = (char)((sample_buffer24[i] >> 16) & 0xFF);
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_32bit(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int          ret;
	unsigned int i;
	long         *sample_buffer32 = (long*)sample_buffer;
	char         *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++){
		data[i*4] = (char)(sample_buffer32[i] & 0xFF);
		data[i*4+1] = (char)((sample_buffer32[i] >> 8) & 0xFF);
		data[i*4+2] = (char)((sample_buffer32[i] >> 16) & 0xFF);
		data[i*4+3] = (char)((sample_buffer32[i] >> 24) & 0xFF);
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}

static int write_audio_float(audio_file *aufile, void *sample_buffer, unsigned int samples)
{
	int           ret;
	unsigned int  i;
	float         *sample_buffer_f = (float*)sample_buffer;
	unsigned char *data = malloc(samples*aufile->bits_per_sample*sizeof(char)/8);

	aufile->samples += samples;

	for (i = 0; i < samples; i++){
		int   exponent,
		      mantissa,
		      negative = 0;
		float in = sample_buffer_f[i];

		data[i*4] = 0;
		data[i*4+1] = 0;
		data[i*4+2] = 0;
		data[i*4+3] = 0;
		if (in == 0.0)
			continue;

		if (in < 0.0){
			in *= -1.0;
			negative = 1;
		}
		in = (float)frexp(in, &exponent);
		exponent += 126;
		in *= (float)0x1000000;
		mantissa = (((int)in) & 0x7FFFFF);

		if (negative)
			data[i*4+3] |= 0x80;

		if (exponent & 0x01)
			data[i*4+2] |= 0x80;

		data[i*4] = mantissa & 0xFF;
		data[i*4+1] = (mantissa >> 8) & 0xFF;
		data[i*4+2] |= (mantissa >> 16) & 0x7F;
		data[i*4+3] |= (exponent >> 1) & 0x7F;
	}

	ret = fwrite(data, samples, aufile->bits_per_sample/8, aufile->sndfile);

	if (data)
		free(data);

	return ret;
}
//void* output_to_PCM(float **input, void *sample_buffer, int channels,
void* output_to_PCM(double **input, void *sample_buffer, int channels,
                    int samples, int format)
{
    int permute[8][8] = {{0}, {0,1}, {0,2,1}, {0,1,2,3}, {0,2,1,3,4}, {0,2,1,5,3,4}, {0,2,1,6,5,3,4}, {0,2,1,7,5,6,3,4}};
	int             i,j,k;

	char            *char_sample_buffer = (char*)sample_buffer;
	short           *short_sample_buffer = (short*)sample_buffer;
	int             *int_sample_buffer = (int*)sample_buffer;
	float           *float_sample_buffer = (float*)sample_buffer;

	/*
	 * Copy output to a standard PCM buffer
	 */
	if (channels > 2 && channels < 9) {
		switch (format) {
		case WAV_FMT_8BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					char_sample_buffer[(i*channels)+j] = input[permute[channels-1][j]][i];
				}
			}
			break;
		case WAV_FMT_16BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					short_sample_buffer[(i*channels)+j] = input[permute[channels-1][j]][i];
				}
			}
			break;
		case WAV_FMT_24BIT:
		case WAV_FMT_32BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					int_sample_buffer[(i*channels)+j] = input[permute[channels-1][j]][i];
				}
			}
			break;
		case WAV_FMT_FLOAT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					float_sample_buffer[(i*channels)+j] = input[permute[channels-1][j]][i];
				}
			}
			break;
		}
	}
	else
	{
		switch (format) {
		case WAV_FMT_8BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					char_sample_buffer[(i*channels)+j] = input[j][i];
				}
			}
			break;
		case WAV_FMT_16BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					short_sample_buffer[(i*channels)+j] = input[j][i];
				}
			}
			break;
		case WAV_FMT_24BIT:
		case WAV_FMT_32BIT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					int_sample_buffer[(i*channels)+j] = input[j][i];
				}
			}
			break;
		case WAV_FMT_FLOAT:
			for (j = 0; j < channels; j++) {
				for (i = 0; i < samples; i++) {
					float_sample_buffer[(i*channels)+j] = input[j][i];
				}
			}
			break;
		}
	}
	return sample_buffer;
}


