/* Set TABS = 4 */
/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggVorbis SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE OggVorbis SOURCE CODE IS (C) COPYRIGHT 1994-2001             *
 * by the XIPHOPHORUS Company http://www.xiph.org/                  *
 *                                                                  *
 ********************************************************************

 function: To provide playback of 16 bit PCM wave data in Win32
           environments from decoded ogg files.

 ********************************************************************/

#include <string.h>
#include <errno.h>
#include "audio.h"
#include "wave_out.h"

#define MAXWAVESIZE     4294967040LU
#define MAX_WAVEBLOCKS    32

// This is modified for USE_WIN_AUDIO - ONLY 2002-02-27


#ifndef _WAVEFORMATEXTENSIBLE_
typedef struct {
    WAVEFORMATEX    outFormat;
    union {
        WORD wValidBitsPerSample;       /* bits of precision  */
        WORD wSamplesPerBlock;          /* valid if wBitsPerSample==0 */
        WORD wReserved;                 /* If neither applies, set to zero. */
    } Samples;
    DWORD           dwChannelMask;      /* which channels are */
                                        /* present in stream  */
    GUID            SubFormat;
} WAVEFORMATEXTENSIBLE, *PWAVEFORMATEXTENSIBLE;
#endif

static const GUID __KSDATAFORMAT_SUBTYPE_IEEE_FLOAT = {WAVE_FORMAT_IEEE_FLOAT, 0x0000, 0x0010, {0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71}};
static const GUID __KSDATAFORMAT_SUBTYPE_PCM = {WAVE_FORMAT_PCM, 0x0000, 0x0010, {0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71}};

static CRITICAL_SECTION  cs;
static HWAVEOUT          dev                    = NULL;
static int               ScheduledBlocks        = 0;
static int               PlayedWaveHeadersCount = 0;          // free index
static WAVEHDR*          PlayedWaveHeaders [MAX_WAVEBLOCKS];

/*
 *  This function registers already played WAVE chunks. Freeing is done by free_memory(),
 */

static void CALLBACK
wave_callback ( HWAVEOUT dev, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2 )
{
	if ( uMsg == WOM_DONE ) {
		EnterCriticalSection ( &cs );
		PlayedWaveHeaders [PlayedWaveHeadersCount++] = (WAVEHDR*) dwParam1;
		LeaveCriticalSection ( &cs );
	}
}


static void
free_memory ( void )
{
	WAVEHDR*  wh;
	HGLOBAL   hg;

	EnterCriticalSection ( &cs );
	wh = PlayedWaveHeaders [--PlayedWaveHeadersCount];
	ScheduledBlocks--;                               // decrease the number of USED blocks
	LeaveCriticalSection ( &cs );

	waveOutUnprepareHeader ( dev, wh, sizeof (WAVEHDR) );

	hg = GlobalHandle ( wh -> lpData );       // Deallocate the buffer memory
	GlobalUnlock (hg);
	GlobalFree   (hg);

	hg = GlobalHandle ( wh );                 // Deallocate the header memory
	GlobalUnlock (hg);
	GlobalFree   (hg);
}


Int
Set_WIN_Params ( FILE_T   dummyFile ,
                 Ldouble  SampleFreq,
                 Uint     BitsPerSample,
                 Uint     Channels,
                 Uint     play_priority )
{
	WAVEFORMATEX          outFormat;
	WAVEFORMATEXTENSIBLE  waveformat;
	UINT          deviceID = WAVE_MAPPER;

	memset(&outFormat, 0, sizeof(WAVEFORMATEX));

	if ( waveOutGetNumDevs () == 0 ) {
		fprintf (stderr, "No audio device present.\n" );
		return -1;
	}

	if (Channels > 2) {
		waveformat.outFormat.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
		waveformat.outFormat.cbSize = sizeof(WAVEFORMATEXTENSIBLE) - sizeof(WAVEFORMATEX);
		if (Channels == 3)
			waveformat.dwChannelMask = 7;
		else if (Channels == 4)
			waveformat.dwChannelMask = 51;
		else if (Channels == 5)
			waveformat.dwChannelMask = 55;
		else if (Channels == 6)
			waveformat.dwChannelMask = 63;
	}
	else {
		waveformat.outFormat.wFormatTag = WAVE_FORMAT_PCM;
	        waveformat.outFormat.cbSize = 0;
		waveformat.dwChannelMask = 0;
	}

	waveformat.outFormat.wBitsPerSample    = BitsPerSample;
	waveformat.outFormat.nChannels         = Channels;
	waveformat.outFormat.nSamplesPerSec    = (unsigned long)(SampleFreq + 0.5);
	waveformat.outFormat.nBlockAlign       = (waveformat.outFormat.wBitsPerSample + 7) / 8 * waveformat.outFormat.nChannels;
	waveformat.outFormat.nAvgBytesPerSec   = waveformat.outFormat.nSamplesPerSec * waveformat.outFormat.nBlockAlign;
	waveformat.Samples.wValidBitsPerSample = waveformat.outFormat.wBitsPerSample;
        waveformat.SubFormat                   = __KSDATAFORMAT_SUBTYPE_PCM;

	switch ( waveOutOpen ( &dev, deviceID, (WAVEFORMATEX *)&waveformat, (DWORD)wave_callback, 0, CALLBACK_FUNCTION ) )
	{
		case MMSYSERR_NOERROR:
			break;
		case MMSYSERR_ALLOCATED:
			fprintf (stderr, "Device is already open.\n");
			return -1;
		case MMSYSERR_BADDEVICEID:
			fprintf (stderr, "The specified device is out of range.\n");
			return -1;
		case MMSYSERR_NODRIVER:
			fprintf (stderr, "There is no audio driver in this system.\n");
			return -1;
		case MMSYSERR_NOMEM:
			fprintf (stderr, "Unable to allocate sound memory\n.");
			return -1;
		case WAVERR_BADFORMAT:
			fprintf (stderr, "This audio format is not supported.\n");
			return -1;
		case WAVERR_SYNC:
			fprintf (stderr, "The device is synchronous.\n");
			return -1;
		default:
			fprintf (stderr, "Unknown media error.\n");
			return -1;
	}

	waveOutReset ( dev );
	InitializeCriticalSection ( &cs );
	switch(play_priority)
	{
		case 0:
			SetPriorityClass ( GetCurrentProcess (), NORMAL_PRIORITY_CLASS );
			break;
		case 1:
			SetPriorityClass ( GetCurrentProcess (), HIGH_PRIORITY_CLASS );
			break;
		case 2:
			SetPriorityClass ( GetCurrentProcess (), REALTIME_PRIORITY_CLASS );
			break;
	}
	return 0;
}


int
WIN_Play_Samples ( const void* data, size_t len )
{
	HGLOBAL    hg;
	HGLOBAL    hg2;
	LPWAVEHDR  wh;
	void*      allocptr;

	do {
		while ( PlayedWaveHeadersCount > 0 )                // free used blocks ...
			free_memory ();

		if ( ScheduledBlocks < sizeof(PlayedWaveHeaders)/sizeof(*PlayedWaveHeaders) ) // wait for a free block ...
			break;
		Sleep (26);
	} while (1);

	if ( (hg2 = GlobalAlloc ( GMEM_MOVEABLE, len )) == NULL ) {  // allocate some memory for a copy of the buffer
		fprintf (stderr, "GlobalAlloc failed." );
		return -1;
	}

	allocptr = GlobalLock (hg2);
	CopyMemory ( allocptr, data, len );                         // Here we can call any modification output functions we want....

	if ( (hg = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (WAVEHDR))) == NULL ) // now make a header and WRITE IT!
		return -1;

	wh                   = GlobalLock (hg);
	wh -> dwBufferLength = len;
	wh -> lpData         = allocptr;

	if ( waveOutPrepareHeader ( dev, wh, sizeof (WAVEHDR)) != MMSYSERR_NOERROR ) {
		GlobalUnlock (hg);
		GlobalFree   (hg);
		return -1;
	}

	if ( waveOutWrite ( dev, wh, sizeof (WAVEHDR)) != MMSYSERR_NOERROR ) {
		GlobalUnlock (hg);
		GlobalFree   (hg);
		return -1;
	}

	EnterCriticalSection ( &cs );
	ScheduledBlocks++;
	LeaveCriticalSection ( &cs );

	return len;
}


int
WIN_Audio_close ( void )
{
	if ( dev != NULL ) {

		while ( ScheduledBlocks > 0 ) {
			Sleep (ScheduledBlocks);
			while ( PlayedWaveHeadersCount > 0 )         // free used blocks ...
				free_memory ();
		}

		waveOutReset (dev);      // reset the device
		waveOutClose (dev);      // close the device
		dev = NULL;
	}

	DeleteCriticalSection ( &cs );
	ScheduledBlocks = 0;
	return 0;
}


/* end of wave_out.c */
